###########################################
package CameraStore::Shell;
###########################################
# Mike Schilli, 2003 (m@perlmeister.com)
###########################################
use warnings;
use strict;

my $ID_REGEX  = qr#^(\d{4})-(\d\d)-(\d\d)
                   _(\d\d):(\d\d):(\d\d)(_\d+)?$#x;
my $TS_REGEX  = qr#^(\d{4}):(\d\d):(\d\d)\s
                   (\d\d):(\d\d):(\d\d)$#x;
my $PIC_REGEX = qr#(\.jpg)$#i;

our %OPTS = ();
our $IMG_FILE_DIR;

use Log::Log4perl qw(:easy);
use Image::Info qw(image_info);
use File::Path;
use File::Basename;
use File::Copy;
use Pod::Usage;

###########################################
sub run {
###########################################
    my($class, %options) = @_;

    my $all_rc = 1;

    if(exists $options{img_file_dir}) {
        $IMG_FILE_DIR = $options{img_file_dir};
    } else {
        LOGDIE "Mandatory parameter 'img_file_dir' missing";
    }

    if(exists $options{cmd_line_options}) {
        %OPTS = %{$options{cmd_line_options}};
    } else {
        LOGDIE "Mandatory parameter 'cmd_line_options' missing";
    }

    my $db = CameraStore->new();

    if($OPTS{v}) {
        print "$CameraStore::VERSION\n";
        return 1;
    }

    if($OPTS{h}) {
        pod2usage(-verbose => 3, -input => CameraStore::locate_me());
        return 1;
    }

    if($OPTS{s}) {
        for($db->search_tag($OPTS{s}, $OPTS{p})) {
            if($OPTS{x}) {
                INFO "Symlinking $_ to ", basename($_);
                unless($OPTS{n}) {
                    symlink $_, basename($_) or warn "Cannot symlink $_ ($!)";
                }
            }
            print "$_\n";
        }
        return 1;
    }

    my $in = get_input_sub();

    while($_ = $in->()) {
    
        my ($id, $dir, $file);
    
            # We are getting either file names or IDs.
        if(-f $_) {
            # It is a file. Add it if so requested.
            ($id, $dir, $file) = file_info($_, \%OPTS);
            if(!$id) {
                ERROR "Couldn't read file meta data";
                $all_rc = undef;
                next;
            }
            if($OPTS{i}) {
                    # file_info will suggest an ID, but it
                    # might be modified by add_file later to
                    # ensure uniqueness.
                ($id, $file) = add_file($db, $id, $dir, $_, $file);
                if(! defined $id) {
                    ERROR "Cannot insert image $_";
                    $all_rc = undef;
                    next;
                }
            }
        } else {
            $id = $_;
        }
    
        unless(defined $id and 
               $id =~ /$ID_REGEX/ and
               $db->_img($id)
              ) {
            $id = 'undef' unless defined $id;
            ERROR "Unknown image id: '$id'. Did you forget -i?";
            $all_rc = undef;
            next;
        }
    
        DEBUG "Image ID=$id";
    
        if($OPTS{t}) {
            my @tags = split /\|/, $OPTS{t};

            for(@tags) {
                INFO("Tagging $id with '$_'");
                $db->add_tag($_, $id);
            }
        } elsif($OPTS{u}) {
            INFO("Untagging $OPTS{u} with $id");
            $db->delete_tag($OPTS{u}, $id);
        } elsif($OPTS{g}) {
            INFO("Grepping $id for '$OPTS{g}'");
            print "$_\n" for
                $db->search_tag($OPTS{g}, 
                                $OPTS{p}, $id);
        } elsif($OPTS{l}) {
            print "$_: ", join(', ',
                $db->list_tags($id)), "\n";
        } elsif($OPTS{i}) {
                # Just imported, we're done.
            next;
        } else {
            pod2usage(-message => "Options error",
                      -verbose => 2, -input => CameraStore::locate_me());
            return undef;
        }
    }

    return $all_rc;
}

###########################################
sub add_file {
###########################################
    my($db, $id, $dir, $ofile, $file) = @_;

    INFO("Importing file $ofile");
    INFO("... with tag $OPTS{tag}") if $OPTS{tag};
    INFO("... with id: $id") if $id;

    if(!-d $dir) {
        unless($OPTS{n}) {
            mkpath($dir) or 
               LOGDIE "Cannot mkpath $dir";
        }
    }

    DEBUG("Adding image $id");
    unless($OPTS{n}) {
        $id = $db->add_image($id, 
                       "$dir/$file", $OPTS{tag}, $OPTS{d}) or return undef;;
    }

    if($id =~ /(_\d+)$/) {
        my $suffix = $1;
        $file =~ s/\./$suffix./;
    }

    DEBUG("Copying $ofile to $dir/$file");
    unless($OPTS{n}) {
        copy($ofile, "$dir/$file") or 
           LOGDIE "$ofile > $dir/$file failed";
    }

    return ($id, $file);
}

###########################################
sub file_info {
###########################################
    my($ofile, $options) = @_;

    my ($suffix) = ($ofile =~ $PIC_REGEX);

    unless($suffix) {
        ERROR "Unknown image type: $ofile";
        return undef;
    }

    my($y, $mo, $d, $h, $mi, $s);

    if($options->{m}) {
        ($s,$mi,$h,$d,$mo,$y) = split /,/,
            sprintf "%02d,%02d,%02d,%02d,%02d,%d", 
            localtime((stat($ofile))[9]);
        $mo++;
        $y += 1900;
    } elsif($options->{o}) {
        ($y, $mo, $d, $h, $mi, $s) = ($options->{o} =~ /(\d+)/g);
    } else {
        ($y, $mo, $d, $h, $mi, $s) = image_date($ofile);
    }

    die "Cannot determine image ID from $ofile" unless defined $s;

    my $stamp = "";
    $stamp = "$y-$mo-${d}_$h:$mi:$s";
    
    my $dir   = "$IMG_FILE_DIR/$y/$mo/$d";
    my $file  = "$h:$mi:$s$suffix";

    DEBUG("file_info: $file: $stamp $dir");
    return($stamp, $dir, $file);
}

###########################################
sub image_date {
###########################################
    my($file) = @_;

    my $info = image_info($file);
    
    if ($info->{error} or
        ! exists $info->{DateTime} or
        $info->{DateTime} !~ $TS_REGEX) {
        WARN "No timestamp from $file";
        return undef;
    }
    return($1, $2, $3, $4, $5, $6);
}

###########################################
sub get_input_sub {
###########################################
    my @items = ();

    if(@ARGV) {
        push @items, @ARGV;
    } else {
        while(<STDIN>) {
            chomp;
            push @items, $_;
        }
    }

    return sub {
        if(@items and -d $items[0]) {
            my $dir = shift @items;
            unshift @items, 
               grep /$PIC_REGEX/, <$dir/*>;
        }
        return shift @items;
    };
}

=head1 NAME

idb - Image database client

=head1 SYNOPSIS

        # Import and tag
    idb -i -t tag [file|dir] ...

        # Same as above, but don't actually modify the database
    idb -n -i -t tag [file|dir] ...

        # Filter files and tag
    ls *.jpg | idb -t tag

        # DB search for tags, print paths
    idb -s search_pattern -p

        # DB search for tags, create symlinks to 
        # all matching images in local directory
    mkdir tmp
    cd tmp
    idb -s search_pattern -p -x

        # Grep for tags in files
    idb -g search_pattern [file|dir] ...

        # List files/tags
    idb -l [file|dir] ...

=head2 Miscellaneous Options

        # -d: Allow timestamp duplicates, resolve by 
        # appending sequence numbers
    idb -i -d -t tag [file|dir]

        # -m: Use modification timestamps of files
        # instead of EXIF headers
    idb -i -m -t tag [file|dir]

        # -o: Override timestamp - specify it on the command line
    idb -i -o 2003-11-02_14:51:49 -t tag [file|dir]

=cut

###########################################
package CameraStore;
###########################################
use warnings;
use strict;

our $VERSION = "0.03";

use Class::DBI;
use File::Spec::Functions;
use Log::Log4perl qw(:easy);

our $SET_DB = ['Main', 
               "dbi:mysql:testidb", 
               '', ''] unless $SET_DB;

###########################################
sub new {       # Constructor
###########################################
    my($class) = @_;
    bless {}, $class;
}

###########################################
sub pod2usage {
###########################################
    Pod::Usage::pod2usage(-input => CameraStore::locate_me());
}
    
###########################################
sub locate_me {
###########################################

    my $subpath  = catfile(split "::", __PACKAGE__) . ".pm";

    DEBUG "Locating the manual for $subpath";

    foreach my $path (@INC) {
        my $fullpath = catfile($path, $subpath);

        if(-e $fullpath) {
            DEBUG "Found: $fullpath";
            return $fullpath;
        }

    }
    DEBUG "Couln't locate $subpath";
    return undef;
}

###########################################
sub _img {      # INTERNAL: Get image by ID
###########################################
    my($self, $stamp) = @_;

    my($img) = CameraStore::IDB::Image->
                 search( stamp => $stamp );

    ERROR("No such ID: '$stamp'") unless
                              defined $img;
    return $img;
}

###########################################
sub _cat {   # INTERNAL: Get category by ID
###########################################
    my($cname) = @_;

    my($cat) = CameraStore::IDB::Category->
            search( name => $cname );

    ERROR "No such tag: $cname" unless
                              defined $cat;
    return $cat;
}

###########################################
sub list_tags {   # Get all tags of one img
###########################################
    my($self, $stamp) = @_;

    my @found = ();

    (my $img = $self->_img($stamp)) or return();

    for my $tag ($img->tags) {
        push @found, $tag->category->name;
    }
  
    return @found;
}

###########################################
sub unique_id {  # Find a unique stamp
###########################################
    my($self, $stamp) = @_;

        # Remove any preexisting suffix
    $stamp =~ s/_\d+$//;
    
    for(my $suffix = 1; ; $suffix++) {
        my $new_stamp = "${stamp}_$suffix";
        if(! CameraStore::IDB::Image->search( 
          stamp => $new_stamp ))  {
            return $new_stamp;
        }
    }
}

###########################################
sub add_image {  # Add new image (plus tag)
###########################################
    my($self, $stamp, $path, $cname, $dupe_ok) = @_;

    if(CameraStore::IDB::Image->search( 
      stamp => $stamp ))  {
        if($dupe_ok) {
            $stamp = $self->unique_id($stamp);
            INFO "Unique ID: $stamp";
                # Modify the path as well
            if($stamp =~ /(_\d+)$/) {
                my $suffix = $1;
                $path =~ s/\./$suffix./;
            }
            INFO "New path: $path";
        } else {
            ERROR "ID $stamp already exists";
            return undef;
        }
    }

    CameraStore::IDB::Image->create({
                     stamp     => $stamp,
                     path      => $path});

    DEBUG "Created new image with stamp $stamp";

    return $stamp unless $cname; # No tag but ok

    my @tags = split /\|/, $cname;

    for(@tags) {
        if(! $self->add_tag($_, $stamp)) {
            return undef;
        }
    }

    return $stamp;
}

###########################################
sub delete_image {# Remove image (and tags)
###########################################
    my($self, $stamp) = @_;

    (my $img = $self->_img($stamp)) or 
        return undef;

    $img->delete();
}

###########################################
sub add_tag {          # Add a new tag name
###########################################
    my($self, $cname, $stamp) = @_;

    INFO "Adding tag $cname to $stamp";

    (my $img = $self->_img($stamp)) or 
        return undef;

        # Add category by name
    my $cat = CameraStore::IDB::Category->
      find_or_create({name => $cname});

    if(CameraStore::IDB::Tag->search(
                   image    => $img->id,
                   category => $cat->id)) {
        ERROR "$stamp already has $cname";
        return undef;
    }

    DEBUG "Tag doesn't exist yet, adding";

        # Add image/cat link to tags table
    $cat->add_to_tags({image => $img->id});
}

###########################################
sub delete_tag {       # Take tag off image
###########################################
    my($self, $cname, $stamp) = @_;

    INFO "Strip $cname from $stamp";

    (my $img = $self->_img($stamp)) or 
        return undef;

    (my $cat = _cat($cname)) or 
        return undef;

    my($tag)=CameraStore::IDB::Tag->search(
                image    => $img->id,
                category => $cat->id,
              );

    unless($tag) {
        ERROR "No $cname on $stamp";
        return undef;
    }

    $tag->delete();
}

###########################################
sub search_tag {
###########################################
    my($self, $tag, $paths, $stamp) = @_;

    my @matches = CameraStore::IDB::Image->
                       match($tag, $stamp);
    my $field = $paths ? "path" : "stamp";
    return map { $_->$field } @matches;
}

###########################################
package CameraStore::IDB::Image;
###########################################
use base q(CameraStore::IDB);
use warnings;
use strict;

__PACKAGE__->table('images');
__PACKAGE__->columns(
    All => qw(id stamp path));
__PACKAGE__->has_many('tags', 
    'CameraStore::IDB::Tag' => 'image');

__PACKAGE__->set_sql(rawmatch => q{
SELECT DISTINCT images.stamp, images.path
FROM categories, tags, images
WHERE
  categories.name LIKE ? AND
  categories.id = tags.category AND
  images.id = tags.image AND
  images.stamp LIKE ?
});

sub match {
    my ($class, $rex, $stamp) = @_;
    $stamp = "%" unless defined $stamp;
    my $sth = $class->sql_rawmatch;
    $sth->execute($rex, $stamp);
    return $class->sth_to_objects($sth);
}

###########################################
package CameraStore::IDB::Category;
###########################################
use base q(CameraStore::IDB);
use warnings;
use strict;

__PACKAGE__->table('categories');
__PACKAGE__->has_many('tags',
    'CameraStore::IDB::Tag' => 'category');
__PACKAGE__->columns(
    All => qw(id name));

###########################################
package CameraStore::IDB::Tag;
###########################################
use base q(CameraStore::IDB);
use warnings;
use strict;

__PACKAGE__->table('tags');
__PACKAGE__->columns(
    All => qw(id image category));
__PACKAGE__->has_a('category',
             'CameraStore::IDB::Category');
__PACKAGE__->has_a('image',
             'CameraStore::IDB::Image');

###########################################
package CameraStore::IDB;
###########################################
use base q(Class::DBI);
use warnings;
use strict;
my $DB_NAME = (split /:/, $CameraStore::SET_DB->[1])[2];
__PACKAGE__->set_db(@{$CameraStore::SET_DB});
use Class::DBI::DATA::Schema;
__PACKAGE__->run_data_sql unless -f $DB_NAME;

1;

__DATA__
CREATE TABLE IF NOT EXISTS images (
  id    INT AUTO_INCREMENT,
  stamp VARCHAR(32),
  path  VARCHAR(255),
  PRIMARY KEY (id)
);

CREATE TABLE IF NOT EXISTS categories (
  id    INT AUTO_INCREMENT,
  name  VARCHAR(128) UNIQUE,
  PRIMARY KEY (id)
);

CREATE TABLE IF NOT EXISTS tags (
  id       INT AUTO_INCREMENT,
  image    INT,
  category INT,
  PRIMARY KEY (id)
);
